__author__ = "Lukas Leufen"
__date__ = '2019-11-15'


from src import helpers
import logging
import argparse


class ExperimentSetup(object):
    """
    params:
    trainable: Train new model if true, otherwise try to load existing model
    """

    def __init__(self, parser_args, **kwargs):
        self.args = self._set_parser_args(parser_args)
        self.data_path = None
        self.experiment_path = None
        self.experiment_name = None
        self.trainable = None
        self.fraction_of_train = None
        self.use_all_stations_on_all_data_sets = None
        self.network = None
        self.var_all_dict = None
        self.stations = None
        self.variables = None
        self.dimensions = None
        self.interpolate_dim = None
        self.target_dim = None
        self.target_var = None
        self.train_kwargs = None
        self.val_kwargs = None
        self.test_kwargs = None
        self.setup_experiment(**kwargs)

    def _set_param(self, param, value, default=None):
        if default is not None:
            value = value.get(param, default)
        setattr(self, param, value)
        logging.debug(f"set experiment attribute: {param}={value}")

    @staticmethod
    def _set_parser_args(args):
        """
        Transform args to dict if given as argparse.Namespace
        :param args:
        :return:
        """
        if isinstance(args, argparse.Namespace):
            return args.__dict__
        return args

    def setup_experiment(self, **kwargs):

        # set data path of this experiment
        self._set_param("data_path", helpers.prepare_host())

        # set experiment name
        exp_date = self.args.get("experiment_date")
        exp_name, exp_path = helpers.set_experiment_name(experiment_date=exp_date)
        self._set_param("experiment_name", exp_name)
        self._set_param("experiment_path", exp_path)
        helpers.check_path_and_create(self.experiment_path)

        # set if model is trainable
        self._set_param("trainable", kwargs, default=True)

        # set fraction of train
        self._set_param("fraction_of_train", kwargs, default=0.8)

        # use all stations on all data sets (train, val, test)
        self._set_param("use_all_stations_on_all_data_sets", kwargs, default=True)
        self._set_param("network", kwargs, default="AIRBASE")
        self._set_param("var_all_dict", kwargs,
                        default={'o3': 'dma8eu', 'relhum': 'average_values', 'temp': 'maximum', 'u': 'average_values',
                                 'v': 'average_values', 'no': 'dma8eu', 'no2': 'dma8eu', 'cloudcover': 'average_values',
                                 'pblheight': 'maximum'})
        self._set_param("stations", kwargs,
                        default=['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087', 'DEBY052', 'DEBY032', 'DEBW022',
                                 'DEBY004', 'DEBY020', 'DEBW030', 'DEBW037', 'DEBW031', 'DEBW015', 'DEBW073', 'DEBY039',
                                 'DEBW038', 'DEBW081', 'DEBY075', 'DEBW040', 'DEBY053', 'DEBW059', 'DEBW027', 'DEBY072',
                                 'DEBW042', 'DEBW039', 'DEBY001', 'DEBY113', 'DEBY089', 'DEBW024', 'DEBW004', 'DEBY037',
                                 'DEBW056', 'DEBW029', 'DEBY068', 'DEBW010', 'DEBW026', 'DEBY002', 'DEBY079', 'DEBW084',
                                 'DEBY049', 'DEBY031', 'DEBW019', 'DEBW001', 'DEBY063', 'DEBY005', 'DEBW046', 'DEBW103',
                                 'DEBW052', 'DEBW034', 'DEBY088', ])
        self._set_param("variables", kwargs, default=list(self.var_all_dict.keys()))
        self._set_param("dimensions", kwargs, default={'new_index': ['datetime', 'Stations']})
        self._set_param("interpolate_dim", kwargs, default='datetime')
        self._set_param("target_dim", kwargs, default='variables')
        self._set_param("target_var", kwargs, default="o3")
        self._set_param("train_kwargs", kwargs, default={"start": "1997-01-01", "end": "2007-12-31"})
        self._set_param("val_kwargs", kwargs, default={"start": "2008-01-01", "end": "2009-12-31"})
        self._set_param("test_kwargs", kwargs, default={"start": "2010-01-01", "end": "2017-12-31"})
