from src.helpers import TimeTracking
import logging
import time
from src.data_generator import DataGenerator
from src.experiment_setup import ExperimentSetup
import argparse
from typing import Dict, List, Any, Tuple


class run(object):
    """
    basic run class to measure execution time. Either call this class calling it by 'with' or delete the class instance
    after finishing the measurement. The duration result is logged.
    """

    del_by_exit = False

    def __init__(self):
        """
        Starts time tracking automatically and logs as info.
        """
        self.time = TimeTracking()
        logging.info(f"{self.__class__.__name__} started")

    def __del__(self):
        """
        This is the class finalizer. The code is not executed if already called by exit method to prevent duplicated
        logging (__exit__ is always executed before __del__) it this class was used in a with statement.
        """
        if not self.del_by_exit:
            self.time.stop()
            logging.info(f"{self.__class__.__name__} finished after {self.time}")
            self.del_by_exit = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__del__()

    @staticmethod
    def do_stuff(length=2):
        time.sleep(length)


class PreProcessing(run):

    """
    Pre-process your data by using this class. It includes time tracking and uses the experiment setup to look for data
    and stores it if not already in local disk. Further, it provides this data as a generator and checks for valid
    stations (in this context: valid=data available). Finally, it splits the data into valid training, validation and
    testing subsets.
    """

    def __init__(self, experiment_setup: ExperimentSetup):
        super().__init__()
        self.setup = experiment_setup
        self.kwargs = None
        self.valid_stations = []
        self._run()

    def _run(self):
        kwargs = {'start': '1997-01-01', 'end': '2017-12-31', 'limit_nan_fill': 1, 'window_history': 13,
                  'window_lead_time': 3, 'interpolate_method': 'linear',
                  'statistics_per_var': self.setup.var_all_dict, }
        args = self.setup.__dict__
        valid_stations = self.check_valid_stations(args, kwargs, self.setup.stations)
        args = self.update_key(args, "stations", valid_stations)
        data_gen = DataGenerator(**args, **kwargs)
        train, val, test = self.split_train_val_test(data_gen, valid_stations, args, kwargs)
        # print stats of data

    def split_train_val_test(self, data, stations, args, kwargs):
        train_index, val_index, test_index = self.split_set_indices(len(stations), args["fraction_of_training"])
        train = self.create_set_split(stations, args, kwargs, train_index, "train")
        val = self.create_set_split(stations, args, kwargs, val_index, "val")
        test = self.create_set_split(stations, args, kwargs, test_index, "test")
        return train, val, test

    @staticmethod
    def split_set_indices(total_length: int, fraction: float) -> Tuple[slice, slice, slice]:
        """
        create the training, validation and test subset slice indices for given total_length. The test data consists on
        (1-fraction) of total_length (fraction*len:end). Train and validation data therefore are made from fraction of
        total_length (0:fraction*len). Train and validation data is split by the factor 0.8 for train and 0.2 for
        validation.
        :param total_length: list with all objects to split
        :param fraction: ratio between test and union of train/val data
        :return: slices for each subset in the order: train, val, test
        """
        pos_test_split = int(total_length * fraction)
        train_index = slice(0, int(pos_test_split * 0.8))
        val_index = slice(int(pos_test_split * 0.8), pos_test_split)
        test_index = slice(pos_test_split, total_length)
        return train_index, val_index, test_index

    def create_set_split(self, stations, args, kwargs, index_list, set_name):
        if args["use_all_stations_on_all_data_sets"]:
            set_stations = stations
        else:
            set_stations = stations[index_list]
        logging.debug(f"{set_name.capitalize()} stations (len={set_stations}): {set_stations}")
        set_kwargs = self.update_kwargs(args, kwargs, f"{set_name}_kwargs")
        set_stations = self.check_valid_stations(args, set_kwargs, set_stations)
        set_args = self.update_key(args, "stations", set_stations)
        data_set = DataGenerator(**set_args, **set_kwargs)
        return data_set

    @staticmethod
    def update_key(orig_dict: Dict, key: str, value: Any) -> Dict:
        """
        create copy of `orig_dict` and update given key by value, returns a copied dict. The original input dict
        `orig_dict` is not modified by this function.
        :param orig_dict: dictionary with arguments that should be updated
        :param key: the key to update
        :param value: the update itself for given key
        :return: updated dict
        """
        updated = orig_dict.copy()
        updated.update({key: value})
        return updated

    @staticmethod
    def update_kwargs(args: Dict, kwargs: Dict, kwargs_name: str):
        """
        copy kwargs and update kwargs parameters by another dictionary stored in args. Not existing keys in kwargs are
        created, existing keys overwritten.
        :param args: dict with the new kwargs parameters stored with key `kwargs_name`
        :param kwargs: dict to update
        :param kwargs_name: key in `args` to find the updates for `kwargs`
        :return: updated kwargs dict
        """
        kwargs_updated = kwargs.copy()
        if kwargs_name in args.keys() and args[kwargs_name]:
            kwargs_updated.update(args[kwargs_name])
        return kwargs_updated

    @staticmethod
    def check_valid_stations(args: Dict, kwargs: Dict, all_stations: List[str]):
        """
        Check if all given stations in `all_stations` are valid. Valid means, that there is data available for the given
        time range (is included in `kwargs`). The shape and the loading time are logged in debug mode.
        :param args: Dictionary with required parameters for DataGenerator class (`data_path`, `network`, `stations`,
            `variables`, `interpolate_dim`, `target_dim`, `target_var`).
        :param kwargs: positional parameters for the DataGenerator class (e.g. `start`, `interpolate_method`,
            `window_lead_time`).
        :param all_stations: All stations to check.
        :return: Corrected list containing only valid station IDs.
        """
        t_outer = TimeTracking()
        t_inner = TimeTracking(start=False)
        logging.info("check valid stations started")
        valid_stations = []

        # all required arguments of the DataGenerator can be found in args, positional arguments in args and kwargs
        data_gen = DataGenerator(**args, **kwargs)
        for station in all_stations:
            t_inner.run()
            try:
                (history, label) = data_gen[station]
                valid_stations.append(station)
                logging.debug(f"{station}: history_shape = {history.shape}")
                logging.debug(f"{station}: loading time = {t_inner}")
            except AttributeError:
                continue
        logging.info(f"run for {t_outer} to check {len(all_stations)} station(s)")
        return valid_stations


class Training(run):

    def __init__(self, setup):
        super().__init__()
        self.setup = setup


class PostProcessing(run):

    def __init__(self, setup):
        super().__init__()
        self.setup = setup


if __name__ == "__main__":

    formatter = '%(asctime)s - %(levelname)s: %(message)s  [%(filename)s:%(funcName)s:%(lineno)s]'
    logging.basicConfig(format=formatter, level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument('--experiment_date', metavar='--exp_date', type=str, nargs=1, default=None,
                        help="set experiment date as string")
    parser_args = parser.parse_args()
    with run():
        setup = ExperimentSetup(parser_args, stations=['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087'])
        PreProcessing(setup)
