import pytest
import logging
from src.modules import run, PreProcessing
from src.helpers import TimeTracking
import src.helpers
from src.experiment_setup import ExperimentSetup
from src.data_generator import DataGenerator
import re
import mock
import numpy as np


class pytest_regex:
    """Assert that a given string meets some expectations."""

    def __init__(self, pattern, flags=0):
        self._regex = re.compile(pattern, flags)

    def __eq__(self, actual):
        return bool(self._regex.match(actual))

    def __repr__(self):
        return self._regex.pattern


class TestRun:

    def test_enter_exit(self, caplog):
        caplog.set_level(logging.INFO)
        with run() as r:
            assert caplog.record_tuples[-1] == ('root', 20, 'run started')
            assert isinstance(r.time, TimeTracking)
            r.do_stuff(0.1)
        assert caplog.record_tuples[-1] == ('root', 20, pytest_regex(r"run finished after \d+\.\d+s"))

    def test_init_del(self, caplog):
        caplog.set_level(logging.INFO)
        r = run()
        assert caplog.record_tuples[-1] == ('root', 20, 'run started')
        r.do_stuff(0.2)
        del r
        assert caplog.record_tuples[-1] == ('root', 20, pytest_regex(r"run finished after \d+\.\d+s"))


class TestPreProcessing:

    def test_init(self, caplog):
        caplog.set_level(logging.INFO)
        setup = ExperimentSetup({}, stations=['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087'],
                                var_all_dict={'o3': 'dma8eu', 'temp': 'maximum'})
        pre = PreProcessing(setup)
        assert caplog.record_tuples[0] == ('root', 20, 'PreProcessing started')
        assert caplog.record_tuples[1] == ('root', 20, 'check valid stations started')
        assert caplog.record_tuples[-1] == ('root', 20, pytest_regex(r'run for \d+\.\d+s to check 5 station\(s\)'))

    def test_run(self):
        pre_processing = object.__new__(PreProcessing)
        pre_processing.time = TimeTracking()
        pre_processing.setup = ExperimentSetup({}, stations=['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087'],
                                               var_all_dict={'o3': 'dma8eu', 'temp': 'maximum'})
        assert pre_processing._run() is None

    def test_split_train_val_test(self):
        pass

    def test_check_valid_stations(self, caplog):
        caplog.set_level(logging.INFO)
        pre = object.__new__(PreProcessing)
        pre.time = TimeTracking()
        pre.setup = ExperimentSetup({}, stations=['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087'],
                                    var_all_dict={'o3': 'dma8eu', 'temp': 'maximum'})
        kwargs = {'start': '1997-01-01', 'end': '2017-12-31', 'limit_nan_fill': 1, 'window_history': 13,
                  'window_lead_time': 3, 'interpolate_method': 'linear',
                  'statistics_per_var': pre.setup.var_all_dict, }
        valids = pre.check_valid_stations(pre.setup.__dict__, kwargs, pre.setup.stations)
        assert valids == pre.setup.stations
        assert caplog.record_tuples[0] == ('root', 20, 'check valid stations started')
        assert caplog.record_tuples[1] == ('root', 20, pytest_regex(r'run for \d+\.\d+s to check 5 station\(s\)'))

    def test_update_kwargs(self):
        args = {"testName": {"testAttribute": "TestValue", "optional": "2019-11-21"}}
        kwargs = {"testAttribute": "DefaultValue", "defaultAttribute": 3}
        updated = PreProcessing.update_kwargs(args, kwargs, "testName")
        assert updated == {"testAttribute": "TestValue", "defaultAttribute": 3, "optional": "2019-11-21"}
        assert kwargs == {"testAttribute": "DefaultValue", "defaultAttribute": 3}
        args = {"testName": None}
        updated = PreProcessing.update_kwargs(args, kwargs, "testName")
        assert updated == {"testAttribute": "DefaultValue", "defaultAttribute": 3}
        args = {"dummy": "notMeaningful"}
        updated = PreProcessing.update_kwargs(args, kwargs, "testName")
        assert updated == {"testAttribute": "DefaultValue", "defaultAttribute": 3}

    def test_update_key(self):
        orig_dict = {"Test1": 3, "Test2": "4", "test3": [1, 2, 3]}
        f = PreProcessing.update_key
        assert f(orig_dict, "Test2", 4) == {"Test1": 3, "Test2": 4, "test3": [1, 2, 3]}
        assert orig_dict == {"Test1": 3, "Test2": "4", "test3": [1, 2, 3]}
        assert f(orig_dict, "Test3", 4) == {"Test1": 3, "Test2": "4", "test3": [1, 2, 3], "Test3": 4}

    def test_split_set_indices(self):
        dummy_list = list(range(0, 15))
        train, val, test = PreProcessing.split_set_indices(len(dummy_list), 0.9)
        assert dummy_list[train] == list(range(0, 10))
        assert dummy_list[val] == list(range(10, 13))
        assert dummy_list[test] == list(range(13, 15))

    @mock.patch("DataGenerator", return_value=object.__new__(DataGenerator))
    @mock.patch("DataGenerator[station]", return_value=(np.ones(10), np.zeros(10)))
    def test_create_set_split(self):
        stations = ['DEBW107', 'DEBY081', 'DEBW013', 'DEBW076', 'DEBW087']
        pre = object.__new__(PreProcessing)
        pre.setup = ExperimentSetup({}, stations=stations, var_all_dict={'o3': 'dma8eu', 'temp': 'maximum'},
                                    train_kwargs={"start": "2000-01-01", "end": "2007-12-31"})
        kwargs = {'start': '1997-01-01', 'end': '2017-12-31', 'statistics_per_var': pre.setup.var_all_dict, }
        train = pre.create_set_split(stations, pre.setup.__dict__, kwargs, slice(0, 3), "train")
        # stopped here. It is a mess with all the different kwargs, args etc. Restructure the idea of how to implement
        # the data sets. Because there are multiple kwargs declarations and which counts in the end. And there are
        # multiple declarations of  the DataGenerator class. Why this? Is it somehow possible, to select elements from
        # this iterator class. Furthermore the names of the DataPrep class is not distinct, because there is no time
        # range provided in file's name. Given the case, that first to total DataGen is called with a short period for
        # data loading. But then, for the data split (I don't know why this could happen, but it is very likely because
        # osf the current multiple declarations of kwargs arguments) the desired time range exceeds the previou
        # mentioned and short time range. But nevertheless, the file with the short period is loaded and used (because
        # during DataPrep loading, the available range is checked).
